#include <filesystem>
#include <iostream>

#include <fggl/gfx/window.hpp>
#include <fggl/gfx/ogl.hpp>
#include <fggl/gfx/renderer.hpp>
#include <fggl/gfx/shader.hpp>
#include <fggl/gfx/camera.hpp>

#include <fggl/ecs2/ecs.cpp>
#include <fggl/gfx/ecs.hpp>
#include <fggl/gfx/ecs-cam.hpp>

#include <fggl/math/ecs.hpp>

#include <fggl/data/procedural.hpp>
#include <fggl/debug/debug.h>
#include <fggl/data/storage.hpp>

#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>

#include <imgui.h>

constexpr bool showNormals = false;

template <typename T> int sgn(T val) {
	return (T(0) < val) - (val < T(0));
}

// prototype of resource discovery
void discover(std::filesystem::path base) {

	std::vector< std::filesystem::path > contentPacks;
	for ( auto& item : std::filesystem::directory_iterator(base) ) {

		// content pack detection
		if ( std::filesystem::is_directory( item ) ) {
			auto manifest = item.path() / "manifest.yml";
			if ( std::filesystem::exists( manifest ) ) {
				contentPacks.push_back( item.path() );
			}
		}

	}

	// what did we find?
	std::cerr << "found pack(s): " << std::endl;
	for ( auto& pack : contentPacks ) {
		std::cerr << pack << std::endl;
	}

}

void loadShader(fggl::gfx::ShaderCache& cache, const char* name, bool hasGeom) {
	fggl::gfx::ShaderConfig config;
	config.name = name;
	config.vertex = config.name + "_vert.glsl";
	config.fragment = config.name + "_frag.glsl";
	cache.load(config);
}

int main(int argc, char* argv[]) {
	fggl::gfx::Context ctx;

	// build our main window
	fggl::gfx::Window win( fggl::gfx::Input::instance() );
	win.title("FGGL Demo");
	win.fullscreen( true );

	// opengl time
	fggl::gfx::Graphics ogl(win);
	fggl::gfx::MeshRenderer meshRenderer;

	// debug layer
	fggl::debug::DebugUI debug(win);
	debug.visible(true);

	// storage API
	fggl::data::Storage storage;
	discover( storage.resolvePath(fggl::data::Data, "res") );

	fggl::gfx::ShaderCache cache(storage);
	loadShader(cache, "unlit", false);
	loadShader(cache, "phong", false);
	loadShader(cache, "normals", true);

	// create ECS
	fggl::ecs2::World world;

	// load the modules
	world.import<fggl::components::Math>();
	world.import<fggl::components::Gfx>();
	world.import<fggl::components::CameraMod>();

	world.import<fggl::systems::OpenGL>();
	world.import<fggl::systems::CameraInput>();

	world.ecs().set<fggl::components::GfxHacks>({ &win });

	// make camera
	auto camEnt = world.create("camera");
	{
		camEnt.set<fggl::components::Transform>({});
		auto trans = camEnt.get_mut<fggl::components::Transform>();
		trans->origin( glm::vec3(0.0f, 3.0f, 3.0f) );

		camEnt.set<fggl::components::Camera>({});
	}

	auto floorEnt = world.create();
	{
		floorEnt.set<fggl::math::Transform>( {} );

		fggl::data::Mesh mesh = fggl::data::make_quad_xz();
		auto token = meshRenderer.upload( mesh );
		floorEnt.set<fggl::components::GfxToken>( token );
	}

	// bunker prefab
	constexpr float HALF_PI = M_PI / 2.0f;
	constexpr int nSections = 2;
	auto bunkerPrototype = world.prefab("bunker");
	{
		bunkerPrototype.add_owned<fggl::components::Transform>();
		bunkerPrototype.add_owned<fggl::components::GfxMat>();

		// procedural mesh
		fggl::data::Mesh mesh;
		for (int i=-(nSections/2); i<=nSections/2; i++) {
			const auto shapeOffset = glm::vec3( 0.0f, 0.0f, i * 1.0f );

			const auto cubeMat = glm::translate( fggl::math::mat4( 1.0f ) , shapeOffset );
			const auto leftSlope = fggl::math::modelMatrix( 
					glm::vec3(-1.0f, 0.0f, 0.0f) + shapeOffset,
					glm::vec3( 0.0f, -HALF_PI, 0.0f) );
			const auto rightSlope = fggl::math::modelMatrix( 
					glm::vec3( 1.0f, 0.0f, 0.0f) + shapeOffset,
					glm::vec3( 0.0f, HALF_PI, 0.0f) );

			fggl::data::make_cube( mesh, cubeMat );
			fggl::data::make_slope( mesh, leftSlope );
			fggl::data::make_slope( mesh, rightSlope );
		}
		mesh.removeDups();

		// upload mesh to GPU and get a rendering token
		auto token = meshRenderer.upload( mesh );
		token.pipeline = cache.get("phong");

		// add the rendering token to the prototype
		bunkerPrototype.set<fggl::components::GfxToken>( token );
	}

	// create testing bunkers
	int nCubes = 3;
	for ( int i=0; i<nCubes; i++ ) {
		auto bunker = world.create(bunkerPrototype);
		//bunker.add<fggl::components::GfxMat>();

		auto trans = bunker.get_mut<fggl::components::Transform>();
		trans->origin( glm::vec3( i * 5.0f, 0.0f, 0.0f) ); 
	}


	fggl::gfx::Input& input = fggl::gfx::Input::instance();

	bool joystickWindow = true;
	bool gamepadWindow = true;

	float time = 0.0f;
	float dt = 16.0f;
	while( !win.closeRequested() ) {
		input.frame();

		ctx.pollEvents();
		debug.frameStart();

		// update step
		time += dt;

		// imgui joystick debug
		if ( joystickWindow ) {
			ImGui::Begin("Joysticks", &joystickWindow);
			for ( int i=0; i<16; i++ ) {
				bool present = input.hasJoystick(i);
				std::string title = "Joystick " + std::to_string(i);

				if ( ImGui::TreeNode(title.c_str()) ) {
					ImGui::Text("present: %s", present ? "yes" : "no" );

					if ( present ) {
						const fggl::gfx::Joystick& joyData = input.joystick(i);
						ImGui::Text( "%s", joyData.name );
						ImGui::Text( "gamepad: %s", joyData.gamepad ? "yes" : "no" );
						ImGui::Text( "axes: %d, buttons: %d, hats: %d",
								joyData.axisCount,
								joyData.buttonCount,
								joyData.hatCount);

						if (ImGui::TreeNode("axis##2")) {
							// dump data
							for ( int axid = 0; axid < joyData.axisCount; axid++ ) {
								ImGui::Text("axis %d, value: %f", 
										axid, joyData.axes[axid] );
							}
							ImGui::TreePop();
						}

						if (ImGui::TreeNode("buttons##2")) {
							// dump data
							for ( int btnid = 0; btnid < joyData.buttonCount; btnid++ ) {
								ImGui::Text("button %d, value: %s", btnid,
										joyData.buttons[btnid] == GLFW_PRESS
										? "down" : "up" );
							}
							ImGui::TreePop();
						}

						if (ImGui::TreeNode("hats##2")) {
							// dump data
							for ( int btnid = 0; btnid < joyData.hatCount; btnid++ ) {
								ImGui::Text("button %d, value: %d",
										btnid, joyData.hats[btnid] );
							}
							ImGui::TreePop();
						}
					}

					ImGui::TreePop();
					ImGui::Separator();
				}

			}
			ImGui::End();
		}

		// imgui gamepad debug
		if ( gamepadWindow ) {
			ImGui::Begin("GamePad", &gamepadWindow);
			for ( int i=0; i<16; i++ ) {
				std::string title = "GamePad " + std::to_string(i);

				bool present = input.hasJoystick(i);
				if ( ImGui::TreeNode(title.c_str()) ) {
					ImGui::Text("present: %s", present ? "yes" : "no" );

					if ( present ) {

						if ( ImGui::TreeNode("buttons##2") ) {
							for ( auto& btn : fggl::gfx::PadButtons ) {
								auto label = fggl::gfx::PadButtonLabels[ (int) btn ];
								ImGui::Text( "%s: %i %i %i", label.c_str(), 
										input.padDown(i, btn),
										input.padPressed(i, btn),
										input.padReleased(i, btn)
									   );
							}
							ImGui::TreePop();
						}

						if ( ImGui::TreeNode("axes##2") ) {
							for ( auto& axis : fggl::gfx::PadAxes ) {
								auto label = fggl::gfx::PadAxisLabels[ (int) axis ];
								ImGui::Text("%s: %f %f", label.c_str(),
										input.padAxis(i, axis),
										input.padAxisDelta(i, axis)
									   );

							}
							ImGui::TreePop();
						}

					}

					ImGui::TreePop();
					ImGui::Separator();
				}

			}
			ImGui::End();
		}

		// render step
		ogl.clear();

		world.tick(dt);

		// render using real shader
		debug.draw();
		win.swap();
	}

	return 0;
}
