#include <fggl/gfx/ecs.hpp>
#include <fggl/gfx/ecs-cam.hpp>

#include <fggl/gfx/window.hpp>

using fggl::math::sgn;

namespace fggl::systems {

	constexpr float ROT_SPEED = 0.05f;
	constexpr float PAN_SPEED = 0.05f;
	constexpr glm::mat4 MAT_IDENTITY(1.0f);

	enum camera_type { cam_free, cam_arcball };
	camera_type cam_mode = cam_free;

	void ProcessCameraInput(fggl::ecs2::impl::iter& it, fggl::components::Transform* t, fggl::components::Camera* c) {
		//FIXME only required because the imput system is not finished
		fggl::gfx::Window* window = it.world().get<fggl::components::GfxHacks>()->window;
		fggl::gfx::Input& input = fggl::gfx::Input::instance();

		//TODO proper input system
		if ( glfwGetKey(window->handle(), GLFW_KEY_F2) == GLFW_PRESS ) {
			cam_mode = cam_free; 
		}
		if ( glfwGetKey(window->handle(), GLFW_KEY_F3) == GLFW_PRESS ) {
			cam_mode = cam_arcball;
		}

		if ( cam_mode == cam_arcball || input.mouseDown( fggl::gfx::MOUSE_2 ) ) {
			CameraArcball(it, t, c);
		} else if ( cam_mode == cam_free ) {
			CameraFree(it, t, c);
		}

		// scroll wheel
		auto& camTransform = t[0];
		auto& camComp = c[0];

		CameraEdgeScroll(input, camTransform, camComp);

		const glm::vec3 dir = ( camTransform.origin() - camComp.target );
		const glm::vec3 forward = glm::normalize( dir );

		glm::vec3 motion(0.0f);
		float delta = (float)input.scrollDeltaY();
		if ( (glm::length( dir ) < 25.0f && delta < 0.0f) || (glm::length( dir ) > 2.5f && delta > 0.0f) )
			motion -= (forward * delta);

		camTransform.origin( camTransform.origin() + motion );
	}

	void CameraArcball(fggl::ecs2::impl::iter& it, fggl::components::Transform* t, fggl::components::Camera* c) {
		const fggl::gfx::Input& input = fggl::gfx::Input::instance();

		auto& camTransform = t[0];
		auto& camComp = c[0];

		// see https://asliceofrendering.com/camera/2019/11/30/ArcballCamera/
		glm::vec4 position(camTransform.origin(), 1.0f);
		glm::vec4 pivot(camComp.target, 1.0f);
		glm::mat4 view = glm::lookAt( camTransform.origin(), camComp.target, camTransform.up() );
		glm::vec3 viewDir = -glm::transpose(view)[2];
		glm::vec3 rightDir = glm::transpose(view)[0];

		float deltaAngleX = ( 2 * M_PI ); 
		float deltaAngleY = ( M_PI );
		float xAngle = ( input.cursorDeltaX() ) * deltaAngleX;
		float yAngle = ( input.cursorDeltaY() ) * deltaAngleY;

		auto cosAngle = glm::dot( viewDir, fggl::math::UP );
		if ( cosAngle * sgn(deltaAngleY) > 0.99f ) {
			deltaAngleY = 0;
		}

		// rotate the camera around the pivot on the first axis
		glm::mat4x4 rotationMatrixX(1.0f);
		rotationMatrixX = glm::rotate( rotationMatrixX, xAngle, fggl::math::UP );
		position = ( rotationMatrixX * ( position - pivot ) ) + pivot;

		// rotate the camera aroud the pivot on the second axis
		glm::mat4x4 rotationMatrixY(1.0f);
		rotationMatrixY = glm::rotate(rotationMatrixY, yAngle, rightDir );
		auto finalPos = ( rotationMatrixY * ( position - pivot ) ) + pivot;

		camTransform.origin( finalPos );
	}


	void CameraFree(fggl::ecs2::impl::iter& it, fggl::components::Transform* t, fggl::components::Camera* c) {

		//FIXME only required because the imput system is not finished
		fggl::gfx::Window* window = it.world().get<fggl::components::GfxHacks>()->window;
		const fggl::gfx::Input& input = fggl::gfx::Input::instance();

		auto& camTransform = t[0];
		auto& camComp = c[0];

		float rotationValue = 0.0f;
		glm::vec3 translation(0.0f);

		// calulate rotation (user input)
		if ( glfwGetKey(window->handle(), GLFW_KEY_Q) == GLFW_PRESS ) {
			rotationValue = ROT_SPEED; 
		} else if ( glfwGetKey(window->handle(), GLFW_KEY_E) == GLFW_PRESS ) {
			rotationValue = -ROT_SPEED;
		}

		// calulate movement (user input)
		if ( glfwGetKey(window->handle(), GLFW_KEY_W) == GLFW_PRESS ) {
			translation -= fggl::math::RIGHT;
		}

		if ( glfwGetKey(window->handle(), GLFW_KEY_S) == GLFW_PRESS ) {
			translation += fggl::math::RIGHT;
		}

		if ( glfwGetKey(window->handle(), GLFW_KEY_D) == GLFW_PRESS ) {
			translation += fggl::math::FORWARD;
		}

		if ( glfwGetKey(window->handle(), GLFW_KEY_A) == GLFW_PRESS ) {
			translation -= fggl::math::FORWARD;
		}

		glm::vec4 position( camTransform.origin(), 1.0f );
		glm::vec4 pivot( camComp.target, 1.0f );

		// apply movement
		if ( translation != glm::vec3(0.0f) ) {
			const auto rotation = (position - pivot);
			const float angle = atan2( rotation.x, rotation.z );
			const auto rotationMat = glm::rotate( MAT_IDENTITY, angle, fggl::math::UP );

			auto deltaMove = (rotationMat * glm::vec4( translation, 1.0f )) * PAN_SPEED;
			deltaMove.w = 0.0f;

			position += deltaMove;
			pivot += deltaMove;
		}

		// apply rotation
		if ( rotationValue != 0.0f ) {
			glm::mat4 rotation = glm::rotate( MAT_IDENTITY, rotationValue, fggl::math::UP );
			position = ( rotation * ( position - pivot ) ) + pivot;
		}

		camTransform.origin( position );
		camComp.target = pivot;
	}

	void CameraEdgeScroll(fggl::gfx::Input& input, fggl::components::Transform& t, fggl::components::Camera& c) {
		auto cursor = input.mousePos();

		float rotationValue = 0.0f;
		glm::vec3 translation(0.0f);

		// calulate movement (user input)
		if ( cursor[1] > 0.95f ) {
			translation += fggl::math::RIGHT;
		}

		if ( cursor[1] < -0.95f ) {
			translation -= fggl::math::RIGHT;
		}

		if ( cursor[0] > 0.95f ) {
			translation += fggl::math::FORWARD;
		}

		if ( cursor[0] < -0.95f ) {
			translation -= fggl::math::FORWARD;
		}

		glm::vec4 position( t.origin(), 1.0f );
		glm::vec4 pivot( c.target, 1.0f );

		// apply movement
		if ( translation != glm::vec3(0.0f) ) {
			const auto rotation = (position - pivot);
			const float angle = atan2( rotation.x, rotation.z );
			const auto rotationMat = glm::rotate( MAT_IDENTITY, angle, fggl::math::UP );

			auto deltaMove = (rotationMat * glm::vec4( translation, 1.0f )) * PAN_SPEED;
			deltaMove.w = 0.0f;

			position += deltaMove;
			pivot += deltaMove;
		}

		// apply rotation
		if ( rotationValue != 0.0f ) {
			glm::mat4 rotation = glm::rotate( MAT_IDENTITY, rotationValue, fggl::math::UP );
			position = ( rotation * ( position - pivot ) ) + pivot;
		}

		t.origin( position );
		c.target = pivot;
	}

};
